@@ -5,18 +5,20 @@ import (
 	"bytes"
 	"encoding/json"
 	"fmt"
-	"golang.org/x/net/http2/hpack"
-	"golang.org/x/net/websocket"
 	"io"
 	"net/http"
+	"regexp"
 	"syscall"
 	"testing"
 	"time"
+
+	"golang.org/x/net/http2/hpack"
+	"golang.org/x/net/websocket"
 )
 
 // TestH1H1PlainGET tests whether simple HTTP/1 GET request works.
 func TestH1H1PlainGET(t *testing.T) {
-	st := newServerTester(nil, t, noopHandler)
+	st := newServerTester(t, options{})
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -33,9 +35,9 @@ func TestH1H1PlainGET(t *testing.T) {
 }
 
 // TestH1H1PlainGETClose tests whether simple HTTP/1 GET request with
-// Connetion: close request header field works.
+// Connection: close request header field works.
 func TestH1H1PlainGETClose(t *testing.T) {
-	st := newServerTester(nil, t, noopHandler)
+	st := newServerTester(t, options{})
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -57,9 +59,12 @@ func TestH1H1PlainGETClose(t *testing.T) {
 // TestH1H1InvalidMethod tests that server rejects invalid method with
 // 501 status code
 func TestH1H1InvalidMethod(t *testing.T) {
-	st := newServerTester(nil, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Errorf("server should not forward this request")
-	})
+	opts := options{
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Errorf("server should not forward this request")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -78,18 +83,16 @@ func TestH1H1InvalidMethod(t *testing.T) {
 // TestH1H1MultipleRequestCL tests that server rejects request which
 // contains multiple Content-Length header fields.
 func TestH1H1MultipleRequestCL(t *testing.T) {
-	st := newServerTester(nil, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Errorf("server should not forward bad request")
-	})
+	opts := options{
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Errorf("server should not forward bad request")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
-	if _, err := io.WriteString(st.conn, fmt.Sprintf(`GET / HTTP/1.1
-Host: %v
-Test-Case: TestH1H1MultipleRequestCL
-Content-Length: 0
-Content-Length: 0
-
-`, st.authority)); err != nil {
+	if _, err := io.WriteString(st.conn, fmt.Sprintf("GET / HTTP/1.1\r\nHost: %v\r\nTest-Case: TestH1H1MultipleRequestCL\r\nContent-Length: 0\r\nContent-Length: 0\r\n\r\n",
+		st.authority)); err != nil {
 		t.Fatalf("Error io.WriteString() = %v", err)
 	}
 
@@ -107,7 +110,7 @@ Content-Length: 0
 // // TestH1H1ConnectFailure tests that server handles the situation that
 // // connection attempt to HTTP/1 backend failed.
 // func TestH1H1ConnectFailure(t *testing.T) {
-// 	st := newServerTester(nil, t, noopHandler)
+// 	st := newServerTester(t, options{})
 // 	defer st.Close()
 
 // 	// shutdown backend server to simulate backend connect failure
@@ -125,9 +128,64 @@ Content-Length: 0
 // 	}
 // }
 
+// TestH1H1AffinityCookie tests that affinity cookie is sent back in
+// cleartext http.
+func TestH1H1AffinityCookie(t *testing.T) {
+	opts := options{
+		args: []string{"--affinity-cookie"},
+	}
+	st := newServerTester(t, opts)
+	defer st.Close()
+
+	res, err := st.http1(requestParam{
+		name: "TestH1H1AffinityCookie",
+	})
+	if err != nil {
+		t.Fatalf("Error st.http1() = %v", err)
+	}
+
+	if got, want := res.status, 200; got != want {
+		t.Errorf("status = %v; want %v", got, want)
+	}
+
+	const pattern = `affinity=[0-9a-f]{8}; Path=/foo/bar`
+	validCookie := regexp.MustCompile(pattern)
+	if got := res.header.Get("Set-Cookie"); !validCookie.MatchString(got) {
+		t.Errorf("Set-Cookie: %v; want pattern %v", got, pattern)
+	}
+}
+
+// TestH1H1AffinityCookieTLS tests that affinity cookie is sent back
+// in https.
+func TestH1H1AffinityCookieTLS(t *testing.T) {
+	opts := options{
+		args: []string{"--alpn-h1", "--affinity-cookie"},
+		tls:  true,
+	}
+	st := newServerTester(t, opts)
+	defer st.Close()
+
+	res, err := st.http1(requestParam{
+		name: "TestH1H1AffinityCookieTLS",
+	})
+	if err != nil {
+		t.Fatalf("Error st.http1() = %v", err)
+	}
+
+	if got, want := res.status, 200; got != want {
+		t.Errorf("status = %v; want %v", got, want)
+	}
+
+	const pattern = `affinity=[0-9a-f]{8}; Path=/foo/bar; Secure`
+	validCookie := regexp.MustCompile(pattern)
+	if got := res.header.Get("Set-Cookie"); !validCookie.MatchString(got) {
+		t.Errorf("Set-Cookie: %v; want pattern %v", got, pattern)
+	}
+}
+
 // TestH1H1GracefulShutdown tests graceful shutdown.
 func TestH1H1GracefulShutdown(t *testing.T) {
-	st := newServerTester(nil, t, noopHandler)
+	st := newServerTester(t, options{})
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -162,15 +220,19 @@ func TestH1H1GracefulShutdown(t *testing.T) {
 	want := io.EOF
 	b := make([]byte, 256)
 	if _, err := st.conn.Read(b); err == nil || err != want {
-		t.Errorf("st.conn.Read(): %v; want %v, %v", err, want)
+		t.Errorf("st.conn.Read(): %v; want %v", err, want)
 	}
 }
 
 // TestH1H1HostRewrite tests that server rewrites Host header field
 func TestH1H1HostRewrite(t *testing.T) {
-	st := newServerTester([]string{"--host-rewrite"}, t, func(w http.ResponseWriter, r *http.Request) {
-		w.Header().Add("request-host", r.Host)
-	})
+	opts := options{
+		args: []string{"--host-rewrite"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			w.Header().Add("request-host", r.Host)
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -190,9 +252,12 @@ func TestH1H1HostRewrite(t *testing.T) {
 // TestH1H1BadHost tests that server rejects request including bad
 // characters in host header field.
 func TestH1H1BadHost(t *testing.T) {
-	st := newServerTester(nil, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Errorf("server should not forward this request")
-	})
+	opts := options{
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Errorf("server should not forward this request")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	if _, err := io.WriteString(st.conn, "GET / HTTP/1.1\r\nTest-Case: TestH1H1HBadHost\r\nHost: foo\"bar\r\n\r\n"); err != nil {
@@ -210,9 +275,12 @@ func TestH1H1BadHost(t *testing.T) {
 // TestH1H1BadAuthority tests that server rejects request including
 // bad characters in authority component of requset URI.
 func TestH1H1BadAuthority(t *testing.T) {
-	st := newServerTester(nil, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Errorf("server should not forward this request")
-	})
+	opts := options{
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Errorf("server should not forward this request")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	if _, err := io.WriteString(st.conn, "GET http://foo\"bar/ HTTP/1.1\r\nTest-Case: TestH1H1HBadAuthority\r\nHost: foobar\r\n\r\n"); err != nil {
@@ -230,9 +298,12 @@ func TestH1H1BadAuthority(t *testing.T) {
 // TestH1H1BadScheme tests that server rejects request including
 // bad characters in scheme component of requset URI.
 func TestH1H1BadScheme(t *testing.T) {
-	st := newServerTester(nil, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Errorf("server should not forward this request")
-	})
+	opts := options{
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Errorf("server should not forward this request")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	if _, err := io.WriteString(st.conn, "GET http*://example.com/ HTTP/1.1\r\nTest-Case: TestH1H1HBadScheme\r\nHost: example.com\r\n\r\n"); err != nil {
@@ -250,9 +321,12 @@ func TestH1H1BadScheme(t *testing.T) {
 // TestH1H1HTTP10 tests that server can accept HTTP/1.0 request
 // without Host header field
 func TestH1H1HTTP10(t *testing.T) {
-	st := newServerTester(nil, t, func(w http.ResponseWriter, r *http.Request) {
-		w.Header().Add("request-host", r.Host)
-	})
+	opts := options{
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			w.Header().Add("request-host", r.Host)
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	if _, err := io.WriteString(st.conn, "GET / HTTP/1.0\r\nTest-Case: TestH1H1HTTP10\r\n\r\n"); err != nil {
@@ -276,9 +350,12 @@ func TestH1H1HTTP10(t *testing.T) {
 // field using actual backend server even if --no-http-rewrite is
 // used.
 func TestH1H1HTTP10NoHostRewrite(t *testing.T) {
-	st := newServerTester(nil, t, func(w http.ResponseWriter, r *http.Request) {
-		w.Header().Add("request-host", r.Host)
-	})
+	opts := options{
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			w.Header().Add("request-host", r.Host)
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	if _, err := io.WriteString(st.conn, "GET / HTTP/1.0\r\nTest-Case: TestH1H1HTTP10NoHostRewrite\r\n\r\n"); err != nil {
@@ -301,21 +378,24 @@ func TestH1H1HTTP10NoHostRewrite(t *testing.T) {
 // TestH1H1RequestTrailer tests request trailer part is forwarded to
 // backend.
 func TestH1H1RequestTrailer(t *testing.T) {
-	st := newServerTester(nil, t, func(w http.ResponseWriter, r *http.Request) {
-		buf := make([]byte, 4096)
-		for {
-			_, err := r.Body.Read(buf)
-			if err == io.EOF {
-				break
+	opts := options{
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			buf := make([]byte, 4096)
+			for {
+				_, err := r.Body.Read(buf)
+				if err == io.EOF {
+					break
+				}
+				if err != nil {
+					t.Fatalf("r.Body.Read() = %v", err)
+				}
 			}
-			if err != nil {
-				t.Fatalf("r.Body.Read() = %v", err)
+			if got, want := r.Trailer.Get("foo"), "bar"; got != want {
+				t.Errorf("r.Trailer.Get(foo): %v; want %v", got, want)
 			}
-		}
-		if got, want := r.Trailer.Get("foo"), "bar"; got != want {
-			t.Errorf("r.Trailer.Get(foo): %v; want %v", got, want)
-		}
-	})
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -339,9 +419,13 @@ func TestH1H1HeaderFieldBufferPath(t *testing.T) {
 	// The value 100 is chosen so that sum of header fields bytes
 	// does not exceed it.  We use > 100 bytes URI to exceed this
 	// limit.
-	st := newServerTester([]string{"--request-header-field-buffer=100"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatal("execution path should not be here")
-	})
+	opts := options{
+		args: []string{"--request-header-field-buffer=100"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatal("execution path should not be here")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -359,9 +443,13 @@ func TestH1H1HeaderFieldBufferPath(t *testing.T) {
 // TestH1H1HeaderFieldBuffer tests that request with header fields
 // larger than configured buffer size is rejected.
 func TestH1H1HeaderFieldBuffer(t *testing.T) {
-	st := newServerTester([]string{"--request-header-field-buffer=10"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatal("execution path should not be here")
-	})
+	opts := options{
+		args: []string{"--request-header-field-buffer=10"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatal("execution path should not be here")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -378,9 +466,13 @@ func TestH1H1HeaderFieldBuffer(t *testing.T) {
 // TestH1H1HeaderFields tests that request with header fields more
 // than configured number is rejected.
 func TestH1H1HeaderFields(t *testing.T) {
-	st := newServerTester([]string{"--max-request-header-fields=1"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatal("execution path should not be here")
-	})
+	opts := options{
+		args: []string{"--max-request-header-fields=1"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatal("execution path should not be here")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -401,9 +493,12 @@ func TestH1H1HeaderFields(t *testing.T) {
 
 // TestH1H1Websocket tests that HTTP Upgrade to WebSocket works.
 func TestH1H1Websocket(t *testing.T) {
-	st := newServerTesterHandler(nil, t, websocket.Handler(func(ws *websocket.Conn) {
-		io.Copy(ws, ws)
-	}))
+	opts := options{
+		handler: websocket.Handler(func(ws *websocket.Conn) {
+			io.Copy(ws, ws)
+		}).ServeHTTP,
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	content := []byte("hello world")
@@ -422,11 +517,15 @@ func TestH1H1Websocket(t *testing.T) {
 // TestH1H1ReqPhaseSetHeader tests mruby request phase hook
 // modifies request header fields.
 func TestH1H1ReqPhaseSetHeader(t *testing.T) {
-	st := newServerTester([]string{"--mruby-file=" + testDir + "/req-set-header.rb"}, t, func(w http.ResponseWriter, r *http.Request) {
-		if got, want := r.Header.Get("User-Agent"), "mruby"; got != want {
-			t.Errorf("User-Agent = %v; want %v", got, want)
-		}
-	})
+	opts := options{
+		args: []string{"--mruby-file=" + testDir + "/req-set-header.rb"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			if got, want := r.Header.Get("User-Agent"), "mruby"; got != want {
+				t.Errorf("User-Agent = %v; want %v", got, want)
+			}
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -444,9 +543,13 @@ func TestH1H1ReqPhaseSetHeader(t *testing.T) {
 // TestH1H1ReqPhaseReturn tests mruby request phase hook returns
 // custom response.
 func TestH1H1ReqPhaseReturn(t *testing.T) {
-	st := newServerTester([]string{"--mruby-file=" + testDir + "/req-return.rb"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatalf("request should not be forwarded")
-	})
+	opts := options{
+		args: []string{"--mruby-file=" + testDir + "/req-return.rb"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatalf("request should not be forwarded")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -480,7 +583,10 @@ func TestH1H1ReqPhaseReturn(t *testing.T) {
 // TestH1H1RespPhaseSetHeader tests mruby response phase hook modifies
 // response header fields.
 func TestH1H1RespPhaseSetHeader(t *testing.T) {
-	st := newServerTester([]string{"--mruby-file=" + testDir + "/resp-set-header.rb"}, t, noopHandler)
+	opts := options{
+		args: []string{"--mruby-file=" + testDir + "/resp-set-header.rb"},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -502,7 +608,10 @@ func TestH1H1RespPhaseSetHeader(t *testing.T) {
 // TestH1H1RespPhaseReturn tests mruby response phase hook returns
 // custom response.
 func TestH1H1RespPhaseReturn(t *testing.T) {
-	st := newServerTester([]string{"--mruby-file=" + testDir + "/resp-return.rb"}, t, noopHandler)
+	opts := options{
+		args: []string{"--mruby-file=" + testDir + "/resp-return.rb"},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -536,7 +645,10 @@ func TestH1H1RespPhaseReturn(t *testing.T) {
 // TestH1H1HTTPSRedirect tests that the request to the backend which
 // requires TLS is redirected to https URI.
 func TestH1H1HTTPSRedirect(t *testing.T) {
-	st := newServerTester([]string{"--redirect-if-not-tls"}, t, noopHandler)
+	opts := options{
+		args: []string{"--redirect-if-not-tls"},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -557,7 +669,13 @@ func TestH1H1HTTPSRedirect(t *testing.T) {
 // TestH1H1HTTPSRedirectPort tests that the request to the backend
 // which requires TLS is redirected to https URI with given port.
 func TestH1H1HTTPSRedirectPort(t *testing.T) {
-	st := newServerTester([]string{"--redirect-if-not-tls", "--redirect-https-port=8443"}, t, noopHandler)
+	opts := options{
+		args: []string{
+			"--redirect-if-not-tls",
+			"--redirect-https-port=8443",
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -576,10 +694,42 @@ func TestH1H1HTTPSRedirectPort(t *testing.T) {
 	}
 }
 
+// TestH1H1POSTRequests tests that server can handle 2 requests with
+// request body.
+func TestH1H1POSTRequests(t *testing.T) {
+	st := newServerTester(t, options{})
+	defer st.Close()
+
+	res, err := st.http1(requestParam{
+		name: "TestH1H1POSTRequestsNo1",
+		body: make([]byte, 1),
+	})
+	if err != nil {
+		t.Fatalf("Error st.http1() = %v", err)
+	}
+	if got, want := res.status, 200; got != want {
+		t.Errorf("res.status: %v; want %v", got, want)
+	}
+
+	res, err = st.http1(requestParam{
+		name: "TestH1H1POSTRequestsNo2",
+		body: make([]byte, 65536),
+	})
+	if err != nil {
+		t.Fatalf("Error st.http1() = %v", err)
+	}
+	if got, want := res.status, 200; got != want {
+		t.Errorf("res.status: %v; want %v", got, want)
+	}
+}
+
 // // TestH1H2ConnectFailure tests that server handles the situation that
 // // connection attempt to HTTP/2 backend failed.
 // func TestH1H2ConnectFailure(t *testing.T) {
-// 	st := newServerTester([]string{"--http2-bridge"}, t, noopHandler)
+// 	opts := options{
+// 		args: []string{"--http2-bridge"},
+// 	}
+// 	st := newServerTester(t, opts)
 // 	defer st.Close()
 
 // 	// simulate backend connect attempt failure
@@ -600,9 +750,13 @@ func TestH1H1HTTPSRedirectPort(t *testing.T) {
 // TestH1H2NoHost tests that server rejects request without Host
 // header field for HTTP/2 backend.
 func TestH1H2NoHost(t *testing.T) {
-	st := newServerTester([]string{"--http2-bridge"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Errorf("server should not forward bad request")
-	})
+	opts := options{
+		args: []string{"--http2-bridge"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Errorf("server should not forward bad request")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	// without Host header field, we expect 400 response
@@ -624,9 +778,13 @@ func TestH1H2NoHost(t *testing.T) {
 // TestH1H2HTTP10 tests that server can accept HTTP/1.0 request
 // without Host header field
 func TestH1H2HTTP10(t *testing.T) {
-	st := newServerTester([]string{"--http2-bridge"}, t, func(w http.ResponseWriter, r *http.Request) {
-		w.Header().Add("request-host", r.Host)
-	})
+	opts := options{
+		args: []string{"--http2-bridge"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			w.Header().Add("request-host", r.Host)
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	if _, err := io.WriteString(st.conn, "GET / HTTP/1.0\r\nTest-Case: TestH1H2HTTP10\r\n\r\n"); err != nil {
@@ -650,9 +808,13 @@ func TestH1H2HTTP10(t *testing.T) {
 // field using actual backend server even if --no-http-rewrite is
 // used.
 func TestH1H2HTTP10NoHostRewrite(t *testing.T) {
-	st := newServerTester([]string{"--http2-bridge"}, t, func(w http.ResponseWriter, r *http.Request) {
-		w.Header().Add("request-host", r.Host)
-	})
+	opts := options{
+		args: []string{"--http2-bridge"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			w.Header().Add("request-host", r.Host)
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	if _, err := io.WriteString(st.conn, "GET / HTTP/1.0\r\nTest-Case: TestH1H2HTTP10NoHostRewrite\r\n\r\n"); err != nil {
@@ -677,11 +839,15 @@ func TestH1H2HTTP10NoHostRewrite(t *testing.T) {
 // concatenates crumbled Cookies automatically, so this test is not
 // much effective now.
 func TestH1H2CrumbleCookie(t *testing.T) {
-	st := newServerTester([]string{"--http2-bridge"}, t, func(w http.ResponseWriter, r *http.Request) {
-		if got, want := r.Header.Get("Cookie"), "alpha; bravo; charlie"; got != want {
-			t.Errorf("Cookie: %v; want %v", got, want)
-		}
-	})
+	opts := options{
+		args: []string{"--http2-bridge"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			if got, want := r.Header.Get("Cookie"), "alpha; bravo; charlie"; got != want {
+				t.Errorf("Cookie: %v; want %v", got, want)
+			}
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -701,11 +867,15 @@ func TestH1H2CrumbleCookie(t *testing.T) {
 // TestH1H2GenerateVia tests that server generates Via header field to and
 // from backend server.
 func TestH1H2GenerateVia(t *testing.T) {
-	st := newServerTester([]string{"--http2-bridge"}, t, func(w http.ResponseWriter, r *http.Request) {
-		if got, want := r.Header.Get("Via"), "1.1 nghttpx"; got != want {
-			t.Errorf("Via: %v; want %v", got, want)
-		}
-	})
+	opts := options{
+		args: []string{"--http2-bridge"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			if got, want := r.Header.Get("Via"), "1.1 nghttpx"; got != want {
+				t.Errorf("Via: %v; want %v", got, want)
+			}
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -722,12 +892,16 @@ func TestH1H2GenerateVia(t *testing.T) {
 // TestH1H2AppendVia tests that server adds value to existing Via
 // header field to and from backend server.
 func TestH1H2AppendVia(t *testing.T) {
-	st := newServerTester([]string{"--http2-bridge"}, t, func(w http.ResponseWriter, r *http.Request) {
-		if got, want := r.Header.Get("Via"), "foo, 1.1 nghttpx"; got != want {
-			t.Errorf("Via: %v; want %v", got, want)
-		}
-		w.Header().Add("Via", "bar")
-	})
+	opts := options{
+		args: []string{"--http2-bridge"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			if got, want := r.Header.Get("Via"), "foo, 1.1 nghttpx"; got != want {
+				t.Errorf("Via: %v; want %v", got, want)
+			}
+			w.Header().Add("Via", "bar")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -747,12 +921,16 @@ func TestH1H2AppendVia(t *testing.T) {
 // TestH1H2NoVia tests that server does not add value to existing Via
 // header field to and from backend server.
 func TestH1H2NoVia(t *testing.T) {
-	st := newServerTester([]string{"--http2-bridge", "--no-via"}, t, func(w http.ResponseWriter, r *http.Request) {
-		if got, want := r.Header.Get("Via"), "foo"; got != want {
-			t.Errorf("Via: %v; want %v", got, want)
-		}
-		w.Header().Add("Via", "bar")
-	})
+	opts := options{
+		args: []string{"--http2-bridge", "--no-via"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			if got, want := r.Header.Get("Via"), "foo"; got != want {
+				t.Errorf("Via: %v; want %v", got, want)
+			}
+			w.Header().Add("Via", "bar")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -772,9 +950,16 @@ func TestH1H2NoVia(t *testing.T) {
 // TestH1H2ReqPhaseReturn tests mruby request phase hook returns
 // custom response.
 func TestH1H2ReqPhaseReturn(t *testing.T) {
-	st := newServerTester([]string{"--http2-bridge", "--mruby-file=" + testDir + "/req-return.rb"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatalf("request should not be forwarded")
-	})
+	opts := options{
+		args: []string{
+			"--http2-bridge",
+			"--mruby-file=" + testDir + "/req-return.rb",
+		},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatalf("request should not be forwarded")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -808,7 +993,13 @@ func TestH1H2ReqPhaseReturn(t *testing.T) {
 // TestH1H2RespPhaseReturn tests mruby response phase hook returns
 // custom response.
 func TestH1H2RespPhaseReturn(t *testing.T) {
-	st := newServerTester([]string{"--http2-bridge", "--mruby-file=" + testDir + "/resp-return.rb"}, t, noopHandler)
+	opts := options{
+		args: []string{
+			"--http2-bridge",
+			"--mruby-file=" + testDir + "/resp-return.rb",
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -842,11 +1033,15 @@ func TestH1H2RespPhaseReturn(t *testing.T) {
 // TestH1H2TE tests that "te: trailers" header is forwarded to HTTP/2
 // backend server by stripping other encodings.
 func TestH1H2TE(t *testing.T) {
-	st := newServerTester([]string{"--http2-bridge"}, t, func(w http.ResponseWriter, r *http.Request) {
-		if got, want := r.Header.Get("te"), "trailers"; got != want {
-			t.Errorf("te: %v; want %v", got, want)
-		}
-	})
+	opts := options{
+		args: []string{"--http2-bridge"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			if got, want := r.Header.Get("te"), "trailers"; got != want {
+				t.Errorf("te: %v; want %v", got, want)
+			}
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -866,9 +1061,14 @@ func TestH1H2TE(t *testing.T) {
 // TestH1APIBackendconfig exercise backendconfig API endpoint routine
 // for successful case.
 func TestH1APIBackendconfig(t *testing.T) {
-	st := newServerTesterConnectPort([]string{"-f127.0.0.1,3010;api;no-tls"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatalf("request should not be forwarded")
-	}, 3010)
+	opts := options{
+		args: []string{"-f127.0.0.1,3010;api;no-tls"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatalf("request should not be forwarded")
+		},
+		connectPort: 3010,
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -903,9 +1103,14 @@ backend=127.0.0.1,3011
 // TestH1APIBackendconfigQuery exercise backendconfig API endpoint
 // routine with query.
 func TestH1APIBackendconfigQuery(t *testing.T) {
-	st := newServerTesterConnectPort([]string{"-f127.0.0.1,3010;api;no-tls"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatalf("request should not be forwarded")
-	}, 3010)
+	opts := options{
+		args: []string{"-f127.0.0.1,3010;api;no-tls"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatalf("request should not be forwarded")
+		},
+		connectPort: 3010,
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -940,9 +1145,14 @@ backend=127.0.0.1,3011
 // TestH1APIBackendconfigBadMethod exercise backendconfig API endpoint
 // routine with bad method.
 func TestH1APIBackendconfigBadMethod(t *testing.T) {
-	st := newServerTesterConnectPort([]string{"-f127.0.0.1,3010;api;no-tls"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatalf("request should not be forwarded")
-	}, 3010)
+	opts := options{
+		args: []string{"-f127.0.0.1,3010;api;no-tls"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatalf("request should not be forwarded")
+		},
+		connectPort: 3010,
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -976,9 +1186,14 @@ backend=127.0.0.1,3011
 
 // TestH1APIConfigrevision tests configrevision API.
 func TestH1APIConfigrevision(t *testing.T) {
-	st := newServerTesterConnectPort([]string{"-f127.0.0.1,3010;api;no-tls"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatalf("request should not be forwarded")
-	}, 3010)
+	opts := options{
+		args: []string{"-f127.0.0.1,3010;api;no-tls"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatalf("request should not be forwarded")
+		},
+		connectPort: 3010,
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -1014,9 +1229,14 @@ func TestH1APIConfigrevision(t *testing.T) {
 // TestH1APINotFound exercise backendconfig API endpoint routine when
 // API endpoint is not found.
 func TestH1APINotFound(t *testing.T) {
-	st := newServerTesterConnectPort([]string{"-f127.0.0.1,3010;api;no-tls"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatalf("request should not be forwarded")
-	}, 3010)
+	opts := options{
+		args: []string{"-f127.0.0.1,3010;api;no-tls"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatalf("request should not be forwarded")
+		},
+		connectPort: 3010,
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -1050,9 +1270,14 @@ backend=127.0.0.1,3011
 
 // TestH1Healthmon tests health monitor endpoint.
 func TestH1Healthmon(t *testing.T) {
-	st := newServerTesterConnectPort([]string{"-f127.0.0.1,3011;healthmon;no-tls"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatalf("request should not be forwarded")
-	}, 3011)
+	opts := options{
+		args: []string{"-f127.0.0.1,3011;healthmon;no-tls"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatalf("request should not be forwarded")
+		},
+		connectPort: 3011,
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
 	res, err := st.http1(requestParam{
@@ -1070,17 +1295,17 @@ func TestH1Healthmon(t *testing.T) {
 // TestH1ResponseBeforeRequestEnd tests the situation where response
 // ends before request body finishes.
 func TestH1ResponseBeforeRequestEnd(t *testing.T) {
-	st := newServerTester([]string{"--mruby-file=" + testDir + "/req-return.rb"}, t, func(w http.ResponseWriter, r *http.Request) {
-		t.Fatal("request should not be forwarded")
-	})
+	opts := options{
+		args: []string{"--mruby-file=" + testDir + "/req-return.rb"},
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			t.Fatal("request should not be forwarded")
+		},
+	}
+	st := newServerTester(t, opts)
 	defer st.Close()
 
-	if _, err := io.WriteString(st.conn, fmt.Sprintf(`POST / HTTP/1.1
-Host: %v
-Test-Case: TestH1ResponseBeforeRequestEnd
-Content-Length: 1000000
-
-`, st.authority)); err != nil {
+	if _, err := io.WriteString(st.conn, fmt.Sprintf("POST / HTTP/1.1\r\nHost: %v\r\nTest-Case: TestH1ResponseBeforeRequestEnd\r\nContent-Length: 1000000\r\n\r\n",
+		st.authority)); err != nil {
 		t.Fatalf("Error io.WriteString() = %v", err)
 	}
 
@@ -1093,3 +1318,34 @@ Content-Length: 1000000
 		t.Errorf("status: %v; want %v", got, want)
 	}
 }
+
+// TestH1H1ChunkedEndsPrematurely tests that an HTTP/1.1 request fails
+// if the backend chunked encoded response ends prematurely.
+func TestH1H1ChunkedEndsPrematurely(t *testing.T) {
+	opts := options{
+		handler: func(w http.ResponseWriter, r *http.Request) {
+			hj, ok := w.(http.Hijacker)
+			if !ok {
+				http.Error(w, "Could not hijack the connection", http.StatusInternalServerError)
+				return
+			}
+			conn, bufrw, err := hj.Hijack()
+			if err != nil {
+				http.Error(w, err.Error(), http.StatusInternalServerError)
+				return
+			}
+			defer conn.Close()
+			bufrw.WriteString("HTTP/1.1 200\r\nTransfer-Encoding: chunked\r\n\r\n")
+			bufrw.Flush()
+		},
+	}
+	st := newServerTester(t, opts)
+	defer st.Close()
+
+	_, err := st.http1(requestParam{
+		name: "TestH1H1ChunkedEndsPrematurely",
+	})
+	if err == nil {
+		t.Fatal("st.http1() should fail")
+	}
+}
